@@ -3,6 +3,7 @@ from torch import nn, einsum
 from inspect import isfunction
 from functools import partial
 from itertools import islice, cycle
+from collections import namedtuple
 import torch.nn.functional as F
 
 from math import sqrt
@@ -19,6 +20,10 @@ from se3_transformer_pytorch import SE3Transformer
 from se3_transformer_pytorch.utils import torch_default_dtype
 from en_transformer import EnTransformer
 
+# constants
+
+Logits = namedtuple('Logits', ['distance', 'theta', 'phi', 'omega'])
+
 # helpers
 
 def exists(val):
@@ -547,10 +552,13 @@ class Alphafold2(nn.Module):
         cross_attn_compress_ratio = 1,
         msa_tie_row_attn = False,
         template_attn_depth = 2,
-        num_backbone_atoms = 1,      # number of atoms to reconstitute each residue to, defaults to 3 for C, C-alpha, N
-        predict_coords = False,      # structure module related keyword arguments below
+        num_backbone_atoms = 1,                # number of atoms to reconstitute each residue to, defaults to 3 for C, C-alpha, N
+        predict_angles = False,
+        predict_coords = False,                # structure module related keyword arguments below
+        predict_real_value_distances = False,
+        return_aux_logits = False,
         mds_iters = 5,
-        use_se3_transformer = True,  # uses SE3 Transformer - but if set to false, will use the new E(n)-Transformer
+        use_se3_transformer = True,            # uses SE3 Transformer - but if set to false, will use the new E(n)-Transformer
         structure_module_dim = 4,
         structure_module_depth = 4,
         structure_module_heads = 1,
@@ -579,6 +587,18 @@ class Alphafold2(nn.Module):
         self.template_pos_emb = nn.Embedding(max_seq_len, dim)
         self.template_pos_emb_ax = nn.Embedding(max_seq_len, dim)
 
+        # projection for angles, if needed
+
+        self.predict_angles = predict_angles
+        if predict_angles:
+            self.to_prob_theta = nn.Linear(dim, constants.THETA_BUCKETS)
+            self.to_prob_phi   = nn.Linear(dim, constants.PHI_BUCKETS)
+            self.to_prob_omega = nn.Linear(dim, constants.OMEGA_BUCKETS)
+
+        # when predicting the coordinates, whether to return the other logits, distogram (and optionally, angles)
+
+        self.return_aux_logits = return_aux_logits
+
         # template sidechain encoding
 
         self.use_se3_transformer = use_se3_transformer
@@ -661,6 +681,9 @@ class Alphafold2(nn.Module):
         self.num_backbone_atoms = num_backbone_atoms
         needs_upsample = num_backbone_atoms > 1
 
+        self.predict_real_value_distances = predict_real_value_distances
+        dim_distance_pred = constants.DISTOGRAM_BUCKETS if not predict_real_value_distances else 2   # 2 for predicting mean and standard deviation values of real-value distance
+
         self.to_distogram_logits = nn.Sequential(
             nn.LayerNorm(dim),
             nn.Sequential(
@@ -669,7 +692,7 @@ class Alphafold2(nn.Module):
                 nn.PixelShuffle(num_backbone_atoms),
                 Rearrange('b c h w -> b h w c')
             ) if needs_upsample else nn.Identity(),
-            nn.Linear(dim, constants.DISTOGRAM_BUCKETS)
+            nn.Linear(dim, dim_distance_pred)
         )
 
         # to coordinate output
@@ -715,6 +738,7 @@ class Alphafold2(nn.Module):
         templates_coors = None,
         templates_sidechains = None,
         embedds = None,
+        return_trunk = False
     ):
         n, device = seq.shape[1], seq.device
         n_range = torch.arange(n, device = device)
@@ -846,28 +870,46 @@ class Alphafold2(nn.Module):
         # embeds to distogram
 
         trunk_embeds = (x + rearrange(x, 'b i j d -> b j i d')) * 0.5  # symmetrize
-        distogram_logits = self.to_distogram_logits(trunk_embeds)
+        distance_pred = self.to_distogram_logits(trunk_embeds)
 
-        if not self.predict_coords:
-            return distogram_logits
+        # determine angles, if specified
+
+        ret = distance_pred
+
+        if self.predict_angles:
+            theta_logits = self.to_prob_theta(x)
+            phi_logits = self.to_prob_phi(x)
+            omega_logits = self.to_prob_omega(x)
+
+            ret = Logits(distance_pred, theta_logits, phi_logits, omega_logits)
+
+        if not self.predict_coords or return_trunk:
+            return ret
 
         # prepare mask for backbone coordinates
 
         assert self.num_backbone_atoms > 1, 'must constitute to at least 3 atomic coordinates for backbone'
 
-        if self.num_backbone_atoms >= 3:
-            N_mask, CA_mask, C_mask = scn_backbone_mask(seq, boolean = True, n_aa = self.num_backbone_atoms)
+        N_mask, CA_mask, C_mask = scn_backbone_mask(seq, boolean = True, n_aa = self.num_backbone_atoms)
 
-            cloud_mask = scn_cloud_mask(seq, boolean = True)
-            flat_cloud_mask = rearrange(cloud_mask, 'b l c -> b (l c)')
-            chain_mask = (mask.unsqueeze(-1) * cloud_mask)
-            flat_chain_mask = rearrange(chain_mask, 'b l c -> b (l c)')
+        cloud_mask = scn_cloud_mask(seq, boolean = True)
+        flat_cloud_mask = rearrange(cloud_mask, 'b l c -> b (l c)')
+        chain_mask = (mask.unsqueeze(-1) * cloud_mask)
+        flat_chain_mask = rearrange(chain_mask, 'b l c -> b (l c)')
 
-            mask = rearrange(chain_mask[:, :, :self.num_backbone_atoms], 'b l c -> b (l c)')
+        bb_mask = rearrange(chain_mask[:, :, :self.num_backbone_atoms], 'b l c -> b (l c)')
+        bb_mask_crossed = rearrange(bb_mask, 'b i -> b i ()') * rearrange(bb_mask, 'b j -> b () j')
 
         # structural refinement
 
-        distances, weights = center_distogram_torch(distogram_logits)
+        if self.predict_real_value_distances:
+            distances, distance_std = distance_pred.unbind(dim = -1)
+            weights = (1 / (1 + distance_std)) # could also do a distance_std.sigmoid() here
+        else:
+            distances, weights = center_distogram_torch(distance_pred)
+
+        weights.masked_fill_(bb_mask_crossed, 0.)
+
         coords_3d, _ = MDScaling(distances, 
             weights = weights,
             iters = self.mds_iters,
@@ -898,4 +940,8 @@ class Alphafold2(nn.Module):
                 x, coords = self.structure_module(x, coords, mask = flat_chain_mask)
 
         coords.type(original_dtype)
+
+        if self.return_aux_logits:
+            return coords, ret
+
         return coords
